#!/usr/bin/python
# Copyright (C) 2017 Cumulus Networks, Inc. all rights reserved
#
# This software is subject to the Cumulus Networks End User License
# Agreement available at the following locations:
#
# Internet: https://cumulusnetworks.com/downloads/eula/latest/view/
# Cumulus Linux systems: /usr/share/cumulus/EULA.txt
#
# portvalidator.py -- validate ports.conf and return a list of any lines
#                     containing syntax errors.
#
# The portvalidator class has two public methods for performing validation
# checks on /etc/cumulus/ports.conf files and displaying the results:
#
# import portsyntaxvalidator
#     Validates syntax (lexical grammar) of the ports.conf file line-by-line
#     against what would be legal for *any* ports.conf declaration line.
#     A mode declaration of 'G25' would fail.  Shorthand modes of `4x' or `/4'
#     would validate, but none of `x4', `4/', `4*', or `4*10G' would validate.
#
# import portsyntaxwriteerrors
#     After validating, write a nicely formatted list of invalid lines and
#     line numbers to stdout.
#
import sys
import argparse
import os
import logging
import cumulus.platforms
from validate import Validator, ValidateError

class ArgParseError(RuntimeError) :
    pass

# ----------------------------------------------------------
#
#   v a l i d a t e __ p o r t s __ m o d e
#
# ----------------------------------------------------------

def validate_ports_mode(debug, mode) :
    """
    Validates the right hand side of a single ports.conf line

    [1-MAXINT]=[4x10G|4x25G|2x50G|40G|50G|100G|100G/4|/1|2x|disabled|loopback]

    for syntactic correctness.  It does *not* validate whether the values
    given are possible, correct, or desirable for this platform.
    """
    if debug :
        print 'Mode: %s' % mode
    if len(mode) == 0 :
        reason = 'blank mode field is invalid'
        if debug :
            print 'validate_ports_mode: %s' % reason
        return (-1, reason)

    if mode.startswith((' ', '\t')) :
        reason = 'mode starting with whitespace is invalid'
        if debug :
            print 'validate_ports_mode: %s' % reason
        return (-1, reason)

    stripped_mode = mode.strip()
    if len(stripped_mode) == 0 :
        reason = 'whitespace-only mode is invalid'
        if debug :
            print 'validate_ports_mode: %s' % reason
        return (-1, reason)

    mode = stripped_mode

    vtortok = Validator()

    # check for simple speeds
    speed_check = 'option("1G", "10G", "13G", "25G", "40G", "50G", "100G")'
    # check for ganged ports
    ganged_check = 'option("2x20G", "2x40G", "2x50G", "4x1G", "4x10G", "4x25G")'
    # check for split ports
    split_check = 'option("40G/4", "100G/4")'
    # check for shorthand gang and split ports
    shorthand_check = 'option("1x", "2x", "4x", "/1", "/2", "/4")'
    # check for string options
    string_check = 'option("disable", "disabled", "loopback", "RJ45", "SFP")'

    try :
        vtortok.check(speed_check, mode)
    except ValidateError :
        if debug :
            print 'Speed check for mode %s failed.' % mode
    else :
        if debug :
            print 'validate_ports_mode: Speed check for mode %s passed.' % mode
        return (0, None)

    try :
        vtortok.check(ganged_check, mode)
    except ValidateError :
        if debug :
            print 'Ganged check for mode %s failed.' % mode
    else :
        if debug :
            print 'validate_ports_mode: Ganged check for mode %s passed.' % mode
        return (0, None)

    try :
        vtortok.check(split_check, mode)
    except ValidateError :
        if debug :
            print 'Split check for mode %s failed.' % mode
    else :
        if debug :
            print 'validate_ports_mode: Split check for mode %s passed.' % mode
        return (0, None)

    try :
        vtortok.check(shorthand_check, mode)
    except ValidateError :
        if debug :
            print 'Shorthand check for mode %s failed.' % mode
    else :
        if debug :
            print 'validate_ports_mode: Shorthand check for mode %s passed.' % mode
        return (0, None)

    try :
        vtortok.check(string_check, mode)
    except ValidateError :
        if debug :
            print 'String check for mode %s failed.' % mode
    else :
        if debug :
            print 'validate_ports_mode: String check for mode %s passed.' % mode
        return (0, None)

    # check for simple speeds
    #speed_check = 'option("1G", "10G", "13G", "25G", "40G", "50G", "100G")'
    # check for ganged ports
    #ganged_check = 'option("2x20G", "2x40G", "2x50G", "4x1G", "4x10G", "4x25G")'
    # check for split ports
    #split_check = 'option("40G/4", "100G/4")'
    # check for shorthand gang and split ports
    #shorthand_check = 'option("1x", "2x", "4x", "/1", "/2", "/4")'
    # check for string options
    #string_check = 'option("disable", "disabled", "loopback", "RJ45", "SFP")'

    reason = 'Mode checks for *G, 4x*G, *G/4, *x, /*, disabled all failed!'
    if debug :
        print 'validate_ports_mode: %s' % reason
    return (-1, reason)

# ----------------------------------------------------------
#
#          v a l i d a t e __ p o r t s __ l i n e
#
# ----------------------------------------------------------

def validate_ports_line(debug, line) :
    """
    Validate a line that is expected to be roughly in the format:

       [#][1-MAXINT]=[4x10G|4x25G|2x50G|40G|50G|100G|100G/4|/1|2x|disabled|loopback]

    Comments (starts with # after leading whitespace is stripped, followed
        by anything up to \n) are automatically valid.  The same is *not*
        true of inline comments following uncommented content in the same line.
    Returns a tuple (retval, reason).
        retval: 0 for valid, -1 for invalid
        reason: if retval == -1, a string explaining why it's invalid
    """
    if debug :
        print 'Line: %s' % line
    if len(line) == 0 :
        if debug :
            print 'validate_ports_line: blank line was declared valid'
        return (0, None)

    stripped_line = line.strip()
    if len(stripped_line) == 0 :
        if debug :
            print 'validate_ports_line: whitespace-only line was declared valid'
        return (0, None)

    if stripped_line[0] == '#' :
        if debug :
            print 'validate_ports_line: comment was declared valid'
        return (0, None)

    vtor = Validator()
    portno_check = 'integer(min=1)'
    line_list = stripped_line.split('=')
    length = len(line_list)
    if length != 2 :
        reason = 'malformed line did not look like <port>=<mode>'
        if debug :
            print 'validate_ports_line: %s' % reason
        return (-1, reason)

    # now we have exactly one = sign, i.e at least it's of the form portno=mode
    # Validate the port number (substring before = sign) first.
    if debug :
        print 'Port: %s' % line_list[0]
    try :
        newvalue = vtor.check(portno_check, line_list[0])
    except ValidateError :
        reason = 'Syntax check for port number failed on %s' % line_list[0]
        if debug :
            print 'validate_ports_line: %s' % reason
        return (-1, reason)
    else :
        if debug :
            print 'Syntax check for port number passed, returned %d.' % newvalue

    mode = line_list[1]
    try :
        (newvalue, reason) = validate_ports_mode(debug, mode)
    except Exception, e :
        logging.exception("validate_ports_line: Couldn't validate mode '%s'" % mode)
    else :
        if newvalue < 0 :
            if debug :
                print 'validate_ports_line: %s' % reason
            return (newvalue, reason)
        else :
            if debug :
                print 'Syntax check for port mode passed.'

    return (0, None)

def portsyntaxwriteerrors(retlist, file='/etc/cumulus/ports.conf') :
    """
    Write the list of syntax validator error lines (invalid lines returned in
        (line_num, line, reason) tuples of retlist[]) with line numbers to
        stdout in a nice format.
    """
    print 'Error: %d invalid lines found in %s:' % (len(retlist), file)

    for (line_num, line, reason) in retlist :
        line = line.rstrip('\r\n')
        print "[Line %d]:'%s'" % (line_num, line)
        print "  Invalid because: %s" % reason

def portsyntaxvalidator(debug=False, file='/etc/cumulus/ports.conf') :
    """
    Read `file' into the lines[] list, stripping trailing carriage returns
        and newlines as we go.  Then validate lines one at a time, and for
        those failing validation add them to a list of tuples containing
        (line number, line, reason).  The list of tuples is returned.
    """
    line_num = 1
    retlist = []

    lines = [line.rstrip('\r\n') for line in open(file)]

    for line in lines :
        try :
            (vterror, reason) = validate_ports_line(debug, line)
        except Exception, e :
            logging.exception("This ports.conf line broke the syntax validator: '%s'" % line)
        else :
            if vterror < 0 :
                retlist.append((line_num, line, reason))
        finally :
            line_num += 1

    return retlist
